from django.utils import timezone

from django.contrib.auth import get_user_model
from django.urls import reverse_lazy
from .models import (Course, CourseProfessor, CourseStudent, Lesson,
                     Video, StudentProgress, Unit, Class,
                     CourseCertification, CertificationProcess, Evaluation,
                     CertificateTemplate, )
from ..activities.models import Activity, Answer
from ..activities.serializers import ActivitySerializer, AnswerSerializer
from courses.notes.models import Note
from courses.certification.serializers import (
    CourseCertificationSerializer,
    CertificationProcessSerializer,
    EvaluationSerializer,
)
from courses.workspaces.models import (
    Workspace,
)
from courses.classes.serializers import ClassSerializer
from rest_framework import serializers

from baquara.users.serializers import SimpleUserSerializer
from baquara.pages.models import BaquaraFlatPage


class BaseCourseSerializer(serializers.ModelSerializer):
    professors = serializers.SerializerMethodField('get_professor_name')
    home_thumbnail_url = serializers.SerializerMethodField()
    is_assistant_or_coordinator = serializers.SerializerMethodField()

    class Meta:
        model = Course
        fields = ("id", "slug", "name", "status", "home_thumbnail_url",
                  "start_date", "home_published", "has_started",
                  "min_percent_to_complete", "professors", "is_assistant_or_coordinator",)

    @staticmethod
    def get_professor_name(obj):
        if obj.course_authors.all():
            return [author.get_name() for author in obj.course_authors.all()]
        return ''

    @staticmethod
    def get_home_thumbnail_url(obj):
        if obj.home_thumbnail:
            return obj.home_thumbnail.url
        return ''

    def get_is_assistant_or_coordinator(self, obj):
        if self.context:
            return obj.is_assistant_or_coordinator(self.context['request'].user)


class BaseClassSerializer(serializers.ModelSerializer):

    class Meta:
        model = Class
        fields = ("id", "name", "assistants", "user_can_certificate")


class VideoSerializer(serializers.ModelSerializer):

    class Meta:
        model = Video
        fields = ('id', 'name', 'youtube_id', 'file',)
        read_only_fields = ('file',)

class CourseSerializer(serializers.ModelSerializer):

    intro_video = VideoSerializer(required=False)
    home_thumbnail_url = serializers.SerializerMethodField()
    professors = SimpleUserSerializer(read_only=True, many=True)
    is_user_assistant = serializers.SerializerMethodField()
    is_user_coordinator = serializers.SerializerMethodField()
    is_assistant_or_coordinator = serializers.SerializerMethodField()

    class Meta:
        model = Course
        fields = ("id", "slug", "name", "intro_video", "application", "requirement",
                  "abstract", "structure", "workload", "pronatec", "status",
                  "thumbnail_url", "home_thumbnail_url", "home_position",
                  "start_date", "home_published", "authors_names", "has_started",
                  "min_percent_to_complete", "is_user_assistant", "is_user_coordinator",
                  "is_assistant_or_coordinator", 'professors', )

    @staticmethod
    def get_home_thumbnail_url(obj):
        if obj.home_thumbnail:
            return obj.home_thumbnail.url
        return ''

    def get_is_user_assistant(self, obj):
        return obj.is_course_assistant(self.context['request'].user)

    def get_is_user_coordinator(self, obj):
        return obj.is_course_coordinator(self.context['request'].user)

    def get_is_assistant_or_coordinator(self, obj):
        return obj.is_assistant_or_coordinator(self.context['request'].user)

    def update(self, instance, validated_data):
        intro_video_data = validated_data.pop('intro_video', None)

        course = super(CourseSerializer, self).update(instance, validated_data)

        if intro_video_data:
            intro_video_ser = VideoSerializer(course.intro_video, data=intro_video_data)
            if intro_video_ser.is_valid():
                intro_video = intro_video_ser.save()
            course.intro_video = intro_video
            course.save()

        return course


class CourseStudentSerializer(serializers.ModelSerializer):
    user = SimpleUserSerializer(read_only=True)

    course_finished = serializers.BooleanField()
    can_emmit_receipt = serializers.BooleanField()
    percent_progress = serializers.IntegerField()
    min_percent_to_complete = serializers.IntegerField()

    current_class = BaseClassSerializer(source='get_current_class')
    course = BaseCourseSerializer()
    certificate = CourseCertificationSerializer()
    last_activity = serializers.SerializerMethodField()

    class Meta:
        model = CourseStudent
        fields = ('id', 'user', 'course', 'start_date', 'course_finished',
                  'certificate', 'can_emmit_receipt', 'percent_progress',
                  'current_class', 'min_percent_to_complete', 'last_activity')

    def get_last_activity(self, obj):
        units = []
        [units.extend(lesson.units.all()) for lesson in obj.course.lessons.all()]
        progresses = StudentProgress.objects.filter(unit__in=units, user=obj.user).order_by('-last_access')
        if progresses:
            return progresses[0].last_access
        return None


class ProfileCourseCertificationSerializer(serializers.ModelSerializer):
    course = BaseCourseSerializer()
    approved = serializers.SerializerMethodField()

    class Meta:
        model = CourseCertification
        fields = ('link_hash', 'created_date', 'is_valid', 'processes', 'type',
                  'approved', 'course')

    def get_approved(self, obj):
        return obj.course_student.can_emmit_receipt()


class ProfileClassSerializer(serializers.ModelSerializer):
    course = serializers.StringRelatedField()

    class Meta:
        model = Class
        fields = ('id', 'name', 'course')


class ProfileSerializer(serializers.ModelSerializer):
    certificates = ProfileCourseCertificationSerializer(many=True,
                                                        source="get_certificates")
    classes = ProfileClassSerializer(many=True)
    groups = serializers.SerializerMethodField(read_only=True)

    def get_groups(self, obj):
        groups_name = obj.groups.all().values_list('name', flat=True)
        return ', '.join(groups_name)

    class Meta:
        model = get_user_model()
        fields = ('id', 'username', 'name', 'first_name', 'last_name',
                  'biography', 'image', 'is_profile_filled', 'occupation',
                  'certificates', 'city', 'site', 'occupation', 'classes',
                  'groups', 'social_facebook', 'social_twitter', 'social_instagram',
                  'phone_number', 'email', 'preferred_language')


class CourseThumbSerializer(serializers.ModelSerializer):

    class Meta:
        model = Course
        fields = ("id", "thumbnail", "home_thumbnail")


class LessonThumbSerializer(serializers.ModelSerializer):

    class Meta:
        model = Lesson
        fields = ("id", "thumbnail", "thumbnail_url")


class StudentProgressSerializer(serializers.ModelSerializer):
    complete = serializers.DateTimeField(required=False)
    user = SimpleUserSerializer(read_only=True, required=False)

    class Meta:
        model = StudentProgress
        fields = ('unit', 'complete', 'user',)


class UnitSerializer(serializers.ModelSerializer):
    video = VideoSerializer(required=False, allow_null=True)
    activities = ActivitySerializer(many=True, required=False, allow_null=True)

    class Meta:
        model = Unit
        fields = ('id', 'title', 'video', 'activities', 'side_notes', 'position', 'chat_room',)


class SimpleUnitSerializer(UnitSerializer):
    video = VideoSerializer(required=False, allow_null=True)

    class Meta:
        model = Unit
        fields = ('id', 'title', 'video', 'position',)


class LessonSerializer(serializers.ModelSerializer):

    units = UnitSerializer(many=True)
    is_course_last_lesson = serializers.BooleanField(read_only=True)
    thumbnail_url = serializers.ReadOnlyField()
    release_date = serializers.DateTimeField(required=False, allow_null=True)
    is_published = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = Lesson
        fields = ('id', 'course', 'is_course_last_lesson', 'desc',
                  'name', 'notes', 'position', 'slug', 'status', 'units',
                  'release_date', 'thumbnail_url', 'is_published')

    def get_is_published(self, instance):
        return instance.release_date and instance.release_date <= timezone.now()

    def update(self, instance, validated_data):
        units = self.update_units(self.initial_data.get('units'), instance)

        for old_unit in instance.units.all():
            if old_unit not in units:
                old_unit.delete()
            else:
                new_activities = units[units.index(old_unit)].activities
                if old_unit.activities != new_activities:
                    for activity in old_unit.activities:
                        if activity not in new_activities:
                            activity.delete()

        validated_data.pop('units')
        return super(LessonSerializer, self).update(instance, validated_data)

    def create(self, validated_data):
        # pop validated data to avoid errors creating lesson
        validated_data.pop('units')
        new_lesson = super(LessonSerializer, self).create(validated_data)

        # use initial_data couse validated_data strips units and videos ids
        units_data = self.initial_data.pop('units')
        self.update_units(units_data, new_lesson)

        return new_lesson

    @classmethod
    def update_units(cls, units_data, lesson):
        units = []
        for unit_data in units_data:
            activities_data = unit_data.pop('activities', None)
            unit_data.pop('lesson', None)

            video_data = unit_data.pop('video', None)

            if video_data:
                video_id = video_data.get('id', None)
                if video_id:
                    video = Video.objects.get(id=video_id)
                    video.youtube_id = video_data.get('youtube_id', '')
                    video.name = video_data.get('name', '')
                else:
                    video = Video(**video_data)

                video.save()
            else:
                video = None
            unit_id = unit_data.pop('id', None)
            if unit_id:
                Unit.objects.filter(id=unit_id).update(video=video, **unit_data)
                unit = Unit.objects.get(id=unit_id)
            else:
                unit = Unit(lesson=lesson, video=video, **unit_data)
            unit.save()

            activities = []
            if activities_data:
                for activity_data in activities_data:
                    activity_id = activity_data.pop('id', None)
                    activity, _ = Activity.objects.get_or_create(id=activity_id)

                    comment = activity_data.get('comment', None)
                    data = activity_data.get('data', None)
                    expected = activity_data.get('expected', None)
                    type_act = activity_data.get('type', None)
                    Activity.objects.filter(id=activity.id).update(
                        comment=comment, data=data,
                        expected=expected, type=type_act, unit=unit
                    )

                    activity = Activity.objects.get(id=activity.id)
                    activity.save()
                    activities.append(activity)
            unit.activities.set(activities)
            units.append(unit)
        return units


class SimpleLessonSerializer(LessonSerializer):
    units = SimpleUnitSerializer(many=True)
    next_url = serializers.SerializerMethodField()
    prev_url = serializers.SerializerMethodField()

    class Meta:
        model = Lesson
        fields = '__all__'

    def get_next_url(self, object):
        course = object.course
        user = None
        request = self.context.get("request")
        if request and hasattr(request, "user"):
            user = request.user

        public_lessons = list(course.public_lessons)
        if (len(public_lessons) > 0) and (object in public_lessons) and (object != public_lessons[-1]):
            index = public_lessons.index(object)
            reverse = reverse_lazy('courses_legacy:lesson', args=[course.slug,
                                                                  public_lessons[index + 1].slug])
        else:
            reverse = reverse_lazy('courses_legacy:course_intro', args=[course.slug])
        return reverse

    def get_prev_url(self, object):
        course = object.course
        user = None
        request = self.context.get("request")
        if request and hasattr(request, "user"):
            user = request.user

        public_lessons = list(course.public_lessons)
        if (len(public_lessons) > 0) and (object in public_lessons) and (object != public_lessons[0]):
            index = public_lessons.index(object)
            reverse = reverse_lazy('courses_legacy:lesson', args=[course.slug,
                                                                  public_lessons[index - 1].slug,
                                                                  public_lessons[index - 1].units.count()])
        else:
            if len(public_lessons) == 0:
                reverse = reverse_lazy('courses_legacy:course_intro', args=[course.slug])
            else:
                reverse = reverse_lazy('courses_legacy:lesson', args=[course.slug,
                                                                  public_lessons[0].slug])
        return reverse


class CourseProfessorSerializer(serializers.ModelSerializer):

    user = serializers.PrimaryKeyRelatedField(queryset=get_user_model().objects.all(), required=False)
    user_info = SimpleUserSerializer(source='user', read_only=True)
    course_info = CourseSerializer(source='course', read_only=True)
    get_name = serializers.CharField(read_only=True)
    get_biography = serializers.CharField(read_only=True)
    get_picture_url = serializers.CharField(read_only=True)
    picture = serializers.ReadOnlyField(source='get_picture_url')
    current_user_classes = ClassSerializer(source='get_current_user_classes', read_only=True, many=True)

    class Meta:
        fields = ('id', 'course', 'course_info', 'user', 'name', 'biography', 'picture', 'user_info',
                  'get_name', 'get_biography', 'get_picture_url', 'role', 'current_user_classes',
                  'is_course_author')
        model = CourseProfessor


class FlatpageSerializer(serializers.ModelSerializer):

    class Meta:
        model = BaquaraFlatPage
        fields = '__all__'
